/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * handshake.c: protocol handshake
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pwd.h>

#include "speextalk.h"

/*
 * Appends c as UTF-8 to the string pointed by pp and updates pp. Returns
 * the number of bytes of the UTF-8 code, or 0 if *pp would go beyond e.
 */
static int
add_utf8(unsigned c, unsigned char **pp, unsigned char *e)
{
    int ml, l, i;
    unsigned char *p = *pp;

    ml = e == NULL || e - p > 6 ? 6 : e - p;
    if(ml == 0)
	return(0);
    if(c < 0x80) {
	*p = c;
	(*pp)++;
	return(1);
    }
    for(l = 2; c >= (2U << (5 * ml)); l++);
    if(ml < l)
	return(0);
    *(p++) = (c >> (6 * l - 6)) | ((0xFC << (6 - l)) & 0xFF);
    for(i = 2; i <= l; i++)
	*(p++) = ((c >> 6 * (l - i)) & 0x3F) | 0x80;
    *pp = p;
    return(l);
}

/*
 * Appends l bytes from t to the string pointed by p, updating p, and
 * 0-terminates the result. Truncates before e.
 */
static void
append_lstring(const unsigned char *t, int l,
    unsigned char **p, const unsigned char *e)
{
    int m = e - *p;

    if(m <= l)
	l = m - 1;
    memcpy(*p, t, l);
    *p += l;
    **p = 0;
}

/*
 * Appends the string t to the string pointed by p, updating p, and
 * 0-terminates the result. Truncates before e.
 */
static void
append_string(const unsigned char *t, unsigned char **p, const unsigned char *e)
{
    append_lstring(t, strlen(t), p, e);
}

/*
 * Appends a string representation of an address to the string pointed by p.
 * Truncated before e.
 */
static void
print_address(struct sockaddr *a, socklen_t la,
    unsigned char **p, unsigned char *e)
{
    char node[1024], service[32];
    int m;

    m = e - *p;
    if(getnameinfo(a, la, node, sizeof(node), service, sizeof(service),
	NI_DGRAM) == 0) {
	append_string(node, p, e);
    } else {
	append_string("<unknown>", p, e);
    }
    append_string(" (", p, e);
    if(getnameinfo(a, la, node, sizeof(node), service, sizeof(service),
	NI_DGRAM | NI_NUMERICHOST | NI_NUMERICSERV) == 0) {
	append_string(node, p, e);
    } else {
	append_string("<unknown>", p, e);
    }
    append_string(")", p, e);
}

/*
 * Returns the handshake packet type if the packet is indeed a handshake
 * packet, 0 else.
 */
static int
is_handshake(unsigned char *p, int lp)
{
    return(lp >= 6 && p[0] == 0 ? p[1] : 0);
}

/*
 * Prints a handshake packet to the interface.
 */
static void
print_handshake(struct sockaddr *a, socklen_t la, unsigned char *p, int lp)
{
    unsigned char pbuf[65536], *s = pbuf, *ebuf = pbuf + sizeof(pbuf);
    int t, lc;

    append_string("From ", &s, ebuf);
    if(a == NULL)
	append_string("peer", &s, ebuf);
    else
	print_address(a, la, &s, ebuf);
    append_string(": ", &s, ebuf);
    if((t = is_handshake(p, lp)) != 0) {
	if(t == 1) {
	    append_string("connection request.\n", &s, ebuf);
	} else if(t == 2) {
	    append_string("connection accepted.\n", &s, ebuf);
	} else if(t == 3) {
	    append_string("connection rejected.\n", &s, ebuf);
	} else {
	    append_string("invalid handshake.\n", &s, ebuf);
	}
	lc = (p[4] << 8) + p[5];
	if(lc > 0 && lc <= lp - 6) {
	    append_lstring(p + 6, lc, &s, ebuf);
	    append_string("\n", &s, ebuf);
	}
	interface_print(pbuf);
    } else {
	append_string("not handshake.\n", &s, ebuf);
    }
}

/*
 * Builds and sends a handshake packet of type t. The comment is taken from
 * the global variable (from the command line), or built from the passwd
 * entry if t == 1.
 */
static void
send_handshake(int t, struct sockaddr *a, socklen_t la)
{
    unsigned char packet[MAX_PACKET_SIZE], *p, *p0, *p1;
    const unsigned char *c, *c1;
    wchar_t wc;
    int lwc;

    packet[0] = 0;
    packet[1] = t;
    packet[2] = packet[3] = 0;
    p = p0 = packet + 6;
    p1 = packet + sizeof(packet);
    if(comment == NULL) {
	if(t == 1) {
	    struct passwd *pw;
	    const char *l;
	    const char *a = " wants to talk to you.";

	    if((pw = getpwuid(getuid())) != NULL) {
		l = pw->pw_gecos;
		if(l == NULL)
		    l = pw->pw_name;
		if(l != NULL) {
		    while(p < p1 && *l != 0)
			*(p++) = *(l++);
		    while(p < p1 && *a != 0)
			*(p++) = *(a++);
		}
	    }
	}
    } else {
	c = comment;
	c1 = strchr(c, 0);
	while(*c != 0) {
	    lwc = mbtowc(&wc, c, c1 - c);
	    if(lwc <= 0) {
		lwc = 1;
		wc = L'?';
	    }
	    c += lwc;
	    if(!add_utf8(wc, &p, p1))
		break;
	}
    }
    packet[4] = (p - p0) >> 8;
    packet[5] = (p - p0) & 0xFF;
    while(sendto(sock, packet, p - packet, 0, a, la) < 0) {
	if(errno != EINTR)
	    fatal_system("send");
    }
}

/*
 * Sends a connection request, and waits for answers. Prints handshake
 * answers. Kills the program on reception of a reject packet. Exits loop on
 * reception of a voice packet.
 */
void
start_handshake(void)
{
    int t;

    send_handshake(1, NULL, 0);
    while(1) {
	receive_packet();
	print_handshake(NULL, 0, recv_buffer, l_recv_buffer);
	t = is_handshake(recv_buffer, l_recv_buffer);
	if(t == 3)
	    exit(1);
	else if(l_recv_buffer >= 4 && recv_buffer[0] == 1)
	    break;
    }
}

/*
 * Waits for handshake packets, and asks the user what to do. Exits loop and
 * connects the socket on acception.
 */
void
wait_handshake(void)
{
    struct sockaddr_storage ra;
    socklen_t lra;

    while(1) {
	lra = sizeof(ra);
	l_recv_buffer = recvfrom(sock, recv_buffer, sizeof(recv_buffer), 0,
	    (struct sockaddr *)&ra, &lra);
	if(l_recv_buffer < 0) {
	    if(errno != EINTR)
		fatal_system("recvfrom");
	}
	print_handshake((struct sockaddr *)&ra, lra,
	    recv_buffer, l_recv_buffer);
	if(l_recv_buffer >= 6 && recv_buffer[0] == 0 && recv_buffer[1] == 1) {
	    if(interface_ask("Take it?")) {
		break;
	    } else {
		send_handshake(3, (struct sockaddr *)&ra, lra);
	    }
	} else {
	    printf("invalid packet.\n");
	}
    }
    if(connect(sock, (struct sockaddr *)&ra, lra))
	fatal_system("connect");
    send_handshake(2, NULL, 0);
}


